package jp.gr.java_conf.matsubara.interpreter.ecmascript.type;

import java.util.HashMap;
import java.util.Iterator;
import java.util.Map;

import jp.gr.java_conf.matsubara.expresser.Expresser;
import jp.gr.java_conf.matsubara.expresser.RuntimeData;
import jp.gr.java_conf.matsubara.expresser.StringToId;
import jp.gr.java_conf.matsubara.expresser.exception.ExpException;
import jp.gr.java_conf.matsubara.expresser.scope.IScope;
import jp.gr.java_conf.matsubara.expresser.scope.Scope;
import jp.gr.java_conf.matsubara.expresser.type.ExpArray;
import jp.gr.java_conf.matsubara.expresser.type.ExpObject;
import jp.gr.java_conf.matsubara.expresser.type.ExpObjectMember;
import jp.gr.java_conf.matsubara.expresser.type.IFunction;
import jp.gr.java_conf.matsubara.expresser.type.IScriptFunction;
import jp.gr.java_conf.matsubara.expresser.type.property.Function__apply;
import jp.gr.java_conf.matsubara.expresser.type.property.Function__call;
import jp.gr.java_conf.matsubara.interpreter.ScriptEngine;
import jp.gr.java_conf.matsubara.interpreter.statement.IStatement;
import jp.gr.java_conf.matsubara.interpreter.statement.StatementResult;
import jp.gr.java_conf.matsubara.interpreter.statement.StatementResultNotification;

/**
 * Function^łB<br/>
 * <br/>
 * 쐬: 2005/08/13
 * 
 * @author m.matsubara
 */
public class EcmaFunction extends ExpObject implements IScriptFunction {
    public static IFunction constructor = null;
    private static final Integer THIS_ID = StringToId.toId("this");            //  thisIuWFNgo^ƂID
    private static final Integer ARGUMENTS_ID = StringToId.toId("arguments");    //  argumentsIdIuWFNgo^ƂID
    private static final Integer CALLEE_ID = StringToId.toId("callee");        //  argumentscalleevpeBID

    /**
     * IDXg
     */
    private Integer[] argIds;
    /**
     * ֐ĂяoɌĂяoXe[gg
     */
    private IStatement functionBody;
    /**
     * [JXR[vŕϐȂꍇɎgpXR[v
     */
    private IScope nextScope;
    
    
    /**
     * [JXR[v̐`<br/>
     * ֐ĂяoƂɂ̐`VXR[v쐬
     */
    private Scope defaultLocalScope;

    /**
     * vg^Cv
     */
    private static ExpObject prototype = null;
    
    /**
     * RXgN^
     * @param argIds IDXg
     * @param nextScope ̃XR[v`FC
     * @throws ExpException
     */
    public EcmaFunction(Integer[] argIds, IStatement functionBody, IScope nextScope) throws ExpException {
        super("Function", constructor);
        this.argIds = argIds;
        this.functionBody = functionBody;
        this.nextScope = nextScope;
        setupLocalVariable();
        
        setThisPrototype(getPrototype());
    }

    public Object callFunction(RuntimeData runtime, Object thisObject,
            Object[] arguments) throws ExpException {
        if (functionBody == null)
            return Expresser.UNDEFINED;

        Object result;
        
        IScope backupCurrentScope = runtime.getCurrentScope();
        IScope scope = newLocalScope(runtime);
        runtime.setCurrentScope(scope);
        try {
            //  this IuWFNg̓o^
//            if (thisUse) {
                if (thisObject == Expresser.UNDEFINED)
                    thisObject = this;  //  TODO zgH
                scope.setValue(THIS_ID, thisObject);
//            }
            result = thisObject;

            //  zϐZbgɊi[
            ExpArray argArray = new ExpArray(arguments);
            argArray.putProperty(CALLEE_ID, this);
            scope.setValue(ARGUMENTS_ID, argArray);
            int max = (argIds.length < arguments.length) ? argIds.length : arguments.length;
            try {
                Scope localScope = (Scope)scope;
                for (int idx = 0; idx < max; idx++)
                    localScope.setVariable(argIds[idx], new ExpObjectMember(argArray, idx));
                for (int idx = max; idx < argIds.length; idx++)
                    localScope.setValue(argIds[idx], Expresser.UNDEFINED);
            } catch (ClassCastException cce) {
                throw new ExpException(ScriptEngine.getResString("InternalError", "FunctionStatement#callFunction", "set variable for scope object"), -1);
            }
            
            
            //  ֐{̂s
            StatementResult receive = functionBody.execute(runtime);
            if (receive != null) {
                int statementType = receive.getStatementType();
                if (statementType == StatementResult.ST_RETURN)
                    return receive.getResultValue();
                if (statementType == StatementResult.ST_BREAK)
                    throw new ExpException(ScriptEngine.getResString("NotProcessedBreak"), -1);
                if (statementType == StatementResult.ST_CONTINUE)
                    throw new ExpException(ScriptEngine.getResString("NotProcessedContinue"), -1);
                if (statementType == StatementResult.ST_EXCEPTION)
                    throw new StatementResultNotification("throw exception.", -1, receive);
            }
        } finally {
            runtime.setCurrentScope(backupCurrentScope);
        }
        return result;
    }

    /*
     *  (non-Javadoc)
     * @see jp.gr.java_conf.matsubara.expresser.type.IScriptFunction#newLocalScope(jp.gr.java_conf.matsubara.expresser.RuntimeData)
     */
    public IScope newLocalScope(RuntimeData runtime) {
        Scope newScope = defaultLocalScope.copyScope();
        newScope.setScopeChain(nextScope);
        return newScope;
    }

    /**
     * ֐ɑ΂āAp郍[JϐȂǂĎgp\ɂ܂B
     *
     */
    public void setupLocalVariable() throws ExpException {
        //  X^bN|C^ɑ΂CfbNXo^
        Map variableMap = new HashMap();    //  IdentityHashMap gc
        defaultLocalScope = new Scope();
        
        functionBody.registLocalVariable(variableMap);

        //  this, arguments o^
        defaultLocalScope.defineVariable(THIS_ID);
        defaultLocalScope.defineVariable(ARGUMENTS_ID);
        
        //  var Xe[gg̕ϐo^
        Iterator iterator = variableMap.keySet().iterator();
        while (iterator.hasNext()) {
            Object id = iterator.next();
            defaultLocalScope.defineVariable((Integer)id);
        }
        //  o^
        for (int idx = 0; idx < argIds.length; idx++) {
            defaultLocalScope.defineVariable(argIds[idx]);
        }
        //  ̌ɕϐǉȂ
        defaultLocalScope.fixed();
    }
    

    /**
     * prototype擾܂B
     * @return prototype 
     * @throws ExpException
     */
    public static synchronized ExpObject getPrototype() throws ExpException {
        if (prototype == null) {
            prototype = new ExpObject(ExpObject.getPrototype());
            prototype.putProperty(Function__apply.ID,   Function__apply.getInstance());
            prototype.putProperty(Function__call.ID,    Function__call.getInstance());
        }
        return prototype;
    }
    
}
